| Link in this file before all others; and link without the
| standard startup code
| A global symbol xfs_entry is defined so that other parts of the
| filesystem might be able to find out about the basepage.

.globl __xfs_entry
__xfs_entry:
	jmp _initialize
